/*
 * ContextEnhanceEagerSearch.cc
 *
 *  Created on: 24.04.2014
 *      Author: lukas
 */

#include "context_enhance_eager_search.h"

#include "globals.h"
#include "heuristic.h"
#include "option_parser.h"
#include "successor_generator.h"
#include "g_evaluator.h"
#include "sum_evaluator.h"
#include "plugin.h"
#include <limits>

using namespace std;

ContextEnhanceEagerSearch::ContextEnhanceEagerSearch(const Options &opts) :
		EagerSearch(opts) {

}

void ContextEnhanceEagerSearch::gebeMatrixAus() {

	for (int i = 0; i < interferenceDistances.size(); i++) {
		for (int j = 0; j < interferenceDistances[i].size(); j++) {
			cout << "[" << interferenceDistances[i][j] << "]";
		}
		cout << endl;
	}
	for (int i = 0; i < g_operators.size(); i++) {
		g_operators[i].dump();
	}
	for (int i = 0; i < g_goal.size(); i++) {
		cout << "var = " << g_goal[i].first << endl;
	}

}
void ContextEnhanceEagerSearch::initialize() {
	//wieviele PriorityQueues brauchen wir? Wie groß ist die maximale Interefenz-Distanz
	Timer init_timer;
	int N = initializeInterferenceDistances();
	init_timer.stop();
	cout << "N = " << N << endl;
	cout << "Init time: " << init_timer << endl;
	open_list->init(N);
	//gebeMatrixAus();
	EagerSearch::initialize();
}

const Operator* ContextEnhanceEagerSearch::getOperatorBy(const State &s) {
	return search_space.get_node(s).get_info().creating_operator;
}
void ContextEnhanceEagerSearch::evaluate_in_openlist(int f, bool is_preferred,
		const State *s, int index_t_strich) {
	if (s != 0) {
		//get PrevCreatingOperator by state s (t in paper)
		const Operator *prevCreatingOperator = getOperatorBy(*s);
		int index_t = -2;
		if (prevCreatingOperator != 0) {
			index_t = prevCreatingOperator - &*g_operators.begin();
			int distance = interferenceDistances[index_t][index_t_strich];
			//set index of next queue
			open_list->setQueueIndex(distance);
		}
		open_list->setPrevCreatingOperator(index_t);

	}

	open_list->evaluate(f, is_preferred);
}

int ContextEnhanceEagerSearch::initializeInterferenceDistances() {

// Set up sizes. (HEIGHT x WIDTH)
	for (int i = 0; i < g_operators.size(); ++i) {
		vector<int> inner_distances;
		inner_distances.resize(g_operators.size(), numeric_limits<int>::max());
		interferenceDistances.push_back(inner_distances);
	}

	//Erste Belegung der Matrix, direkte Interferenz
	for (size_t x = 0; x < g_operators.size(); ++x) {
		Operator *op = &g_operators[x];
		for (size_t y = 0; y < g_operators.size(); ++y) {
			Operator *some_op = &g_operators[y];

			if (op == some_op) {
				interferenceDistances[x][y] = 0;
			} else if (interfere(op, some_op)) {
				setMatrixSymmetrically(x, y, 1);
			}
		}
	}
	//Floyd Algorithmus
	for (size_t k = 0; k < g_operators.size(); ++k) {
		for (size_t i = 0; i < g_operators.size(); ++i) {
			for (size_t j = 0; j < g_operators.size(); ++j) {
				if (!(interferenceDistances[i][k] == numeric_limits<int>::max()
					|| interferenceDistances[k][j]== numeric_limits<int>::max()))
					interferenceDistances[i][j] = min(
						 	interferenceDistances[i][j],interferenceDistances[i][k] + interferenceDistances[k][j]);

			}
		}
		//	if (keineUnendlichenDistanzenDrin())
		//	break;
	}

	return getBiggestDistance();

}

bool ContextEnhanceEagerSearch::keineUnendlichenDistanzenDrin() {
	for (int i = 0; i < interferenceDistances.size(); i++) {
		for (int j = 0; j < interferenceDistances[i].size(); j++) {
			if (interferenceDistances[i][j] == g_operators.size())
				return false;
		}
	}
	return true;
}

bool ContextEnhanceEagerSearch::interfere(Operator *op_first,
		Operator *op_second) {

	const vector<PrePost> &pre_post_first = op_first->get_pre_post();
	const vector<Prevail> &prevail_first = op_first->get_prevail();

	const vector<PrePost> &pre_post_second = op_second->get_pre_post();
	const vector<Prevail> &prevail_second = op_second->get_prevail();

//var(eff(t1)) ^ var(pre(t2)) != 0
	for (int i = 0; i < pre_post_first.size(); i++) {
		PrePost prePost = pre_post_first[i];
		for (int j = 0; j < prevail_second.size(); j++) {
			Prevail prevail = prevail_second[j];
			if (prevail.var == prePost.var)
				return true;
		}
		for (int k = 0; k < pre_post_second.size(); k++) {
			PrePost prePost_sec = pre_post_second[k];
			if (prePost_sec.pre != -1 && prePost.var == prePost_sec.var)
				return true;
		}
	}

//var(eff(t2)) ^ var(pre(t1)) != 0
	for (int i = 0; i < pre_post_second.size(); i++) {
		PrePost prePost = pre_post_second[i];
		for (int j = 0; j < prevail_first.size(); j++) {
			Prevail prevail = prevail_first[j];
			if (prevail.var == prePost.var)
				return true;
		}
		for (int k = 0; k < pre_post_first.size(); k++) {
			PrePost prePost_first = pre_post_first[k];
			if (prePost_first.pre != -1 && prePost.var == prePost_first.var)
				return true;
		}
	}

//var(eff(t1)) ^ var(eff(t2)) != 0
	for (int i = 0; i < pre_post_first.size(); i++) {
		PrePost prePost_first = pre_post_first[i];
		for (int j = 0; j < pre_post_second.size(); j++) {
			PrePost prePost_sec = pre_post_second[j];
			if (prePost_first.var == prePost_sec.var)
				return true;
		}

	}

//otherwise false
	return false;
}

void ContextEnhanceEagerSearch::setMatrixSymmetrically(int index_first,
		int index_second, int value) {

	interferenceDistances[index_first][index_second] = value;
	interferenceDistances[index_second][index_first] = value;
}

int ContextEnhanceEagerSearch::getBiggestDistance() {
	int N = 0;
	for (int i = 0; i < interferenceDistances.size(); i++) {
		for (int j = 0; j < interferenceDistances[i].size(); j++) {
			if (interferenceDistances[i][j] != numeric_limits<int>::max()
					&& interferenceDistances[i][j] > N) {
				N = interferenceDistances[i][j];
			}
		}
	}
	return N;
}

static SearchEngine *_parse_context_enhanced(OptionParser &parser) {
	parser.document_synopsis("Greedy search (eager)", "");

	parser.add_list_option<ScalarEvaluator *>("evals", "scalar evaluators");

	SearchEngine::add_options_to_parser(parser);

	Options opts = parser.parse();
	opts.verify_list_non_empty<ScalarEvaluator *>("evals");

	opts.set("pref_only", false); //otherwise compile err

	EagerSearch *engine = 0;
	if (!parser.dry_run()) {
		vector<ScalarEvaluator *> evals = opts.get_list<ScalarEvaluator *>(
				"evals");
		OpenList<StateID> *open;
		open = new ContextEnhanceOpenList<StateID>(evals[0], false);

		opts.set("open", open);
		opts.set("reopen_closed", false);
		opts.set("pathmax", false);
		opts.set("mpd", false);
		ScalarEvaluator *sep = 0;
		opts.set("f_eval", sep);
		engine = new ContextEnhanceEagerSearch(opts);
	}
	return engine;
}

static Plugin<SearchEngine> _plugin("context_enhanced_eager",
		_parse_context_enhanced);
