/*
 * context_enhance_open_list.cc
 *
 *  Created on: 25.04.2014
 *      Author: lukas
 */

#ifdef CONTEXT_ENHANCE_OPEN_LIST_H_

#include "context_enhance_open_list.h"
#include "../scalar_evaluator.h"
#include "../option_parser.h"
#include "standard_scalar_open_list.h"
#include <iostream>
#include <limits>

using namespace std;

template<class Entry>
ContextEnhanceOpenList<Entry>::ContextEnhanceOpenList(const Options &opts)
: OpenList<Entry>(opts.get<bool>("pref_only")),
size(0),
evaluator(opts.get<ScalarEvaluator *>("eval")) {

}

template<class Entry>
ContextEnhanceOpenList<Entry>::ContextEnhanceOpenList(
		ScalarEvaluator *eval, bool preferred_only)
: OpenList<Entry>(preferred_only), size(0), evaluator(eval) {
}

template<class Entry>
void ContextEnhanceOpenList<Entry>::init(int numberOfQueues) {
	//initialize open_lists bzw. inner_lists
	for (int i = 0; i <= numberOfQueues; i++) {
		open_lists.push_back(
				new StandardScalarOpenList<StateID>(evaluator, false));
	}

	this->queueIndex = 0;
	this->index_t = -2; //am Anfang gibt es sozusagen keinen Operator t

}

template<class Entry>
void ContextEnhanceOpenList<Entry>::setPrevCreatingOperator(int index_t) {
	this->index_t = index_t;
}

template<class Entry>
void ContextEnhanceOpenList<Entry>::setQueueIndex(int n) {
	queueIndex = n;
}

template<class Entry>
bool ContextEnhanceOpenList<Entry>::isInnocent() {

	if(index_t != -2) {
		Operator* t = &g_operators[index_t];
		const vector<PrePost> &pre_post_first = t->get_pre_post();

		for (int i = 0; i < pre_post_first.size(); i++) {
			PrePost prePost = pre_post_first[i];
			for (int j = 0; j < g_goal.size(); j++) {
				if(prePost.var == g_goal[j].first && prePost.post == g_goal[j].second) {
					return false;
				}
			}
		}
	} else {
		return false;
	}

	return true;
}

template<class Entry>
int ContextEnhanceOpenList<Entry>::insert(const Entry &entry) {
	//siehe paper

	if(isInnocent()) {
		if(queueIndex != numeric_limits<int>::max()) //nur einfügen, wenn es interferiert ansonsten prune (nicht einfügen)
		if (!open_lists[queueIndex]->is_dead_end()) {
			open_lists[queueIndex]->insert(entry);
			size++;
		}

		/*else{
		 cout << "BLA 1" << endl;
		 open_lists[open_lists.size()-1]->insert(entry);
		 cout << "BLA 2" << endl;
		 size++;
		 }*/
	} else
	if (!open_lists[0]->is_dead_end()) {
		open_lists[0]->insert(entry);
		size++;
	}

	return 1;
}

template<class Entry>
Entry ContextEnhanceOpenList<Entry>::remove_min(std::vector<int> *key) {

	int best = 0;
	for(int i = 0; i < open_lists.size(); i++) {
		if (!open_lists[i]->empty()) {
			best = i;
			break;
		}
	}
	size--;
	return open_lists[best]->remove_min(key);
}

template<class Entry>
void ContextEnhanceOpenList<Entry>::evaluate(int g, bool preferred) {
	/*
	 Treat as a dead end if
	 1. at least one heuristic reliably recognizes it as a dead end, or
	 2. all heuristics unreliably recognize it as a dead end
	 In case 1., the dead end is reliable; in case 2. it is not.
	 */

	dead_end = true;
	dead_end_reliable = false;
	for (size_t i = 0; i < open_lists.size(); i++) {
		open_lists[i]->evaluate(g, preferred);
		if (open_lists[i]->is_dead_end()) {
			if (open_lists[i]->dead_end_is_reliable()) {
				dead_end = true; // Might have been set to false.
				dead_end_reliable = true;
				break;
			}
		} else {
			dead_end = false;
		}
	}
}

template<class Entry>
bool ContextEnhanceOpenList<Entry>::empty() const {
	return size == 0;
}

template<class Entry>
void ContextEnhanceOpenList<Entry>::clear() {
	size = 0;
	for (size_t i = 0; i < open_lists.size(); i++)
	open_lists[i]->clear();
}

template<class Entry>
bool ContextEnhanceOpenList<Entry>::is_dead_end() const {
	return dead_end;
}

template<class Entry>
bool ContextEnhanceOpenList<Entry>::dead_end_is_reliable() const {
	return dead_end_reliable;
}

template<class Entry>
void ContextEnhanceOpenList<Entry>::get_involved_heuristics(
		std::set<Heuristic *> &hset) {
	for (size_t i = 0; i < open_lists.size(); i++)
	open_lists[i]->get_involved_heuristics(hset);
}

#endif
